// import 'source-map-support/register';
import { YGOCardConfig } from './config';
import SQL from 'sql.js';
import { promises as fs } from 'fs';
import _ from 'lodash';
import { YGOProCard } from './YGOProCard';
import { plainToInstance } from 'class-transformer';
import path from 'path';
import {
  StarterPlugin,
  DefinePlugin,
  LifecycleEvents,
  InjectLogger,
  UseCommand,
  PutOption,
  PutArg,
  PutSession,
  CommandLocale,
  Inject,
} from 'koishi-thirdeye';
import { Logger, Random, Session } from 'koishi';
import initSqlJs from 'sql.js';
import * as localeZh from './locales/zh';
import * as localeEn from './locales/en';
import PuppeteerPlugin from 'koishi-plugin-puppeteer';

export * from './config';

const commonFilter =
  'datas.type & 0x4000 = 0 and (datas.alias = 0 or datas.id - datas.alias > 10)';

@DefinePlugin({ name: 'ygocard', schema: YGOCardConfig })
export default class YGOCardPlugin
  extends StarterPlugin(YGOCardConfig)
  implements LifecycleEvents
{
  @InjectLogger()
  private logger: Logger;

  @Inject()
  private puppeteer: PuppeteerPlugin;

  private dbs: SQL.Database[] = [];
  private querySQL(db: SQL.Database, sql: string, params: any = {}) {
    const statement = db.prepare(sql);
    statement.bind(params);
    const results: any[] = [];
    while (statement.step()) {
      results.push(statement.getAsObject());
    }
    statement.free();
    return results;
  }
  private queryInDB(value: string, db: SQL.Database) {
    const possibleValueNumber = parseInt(value) || 0;
    const results = this.querySQL(
      db,
      `select datas.*,texts.name,texts.desc from datas,texts where datas.id = texts.id and ${commonFilter} and (datas.id = $num or texts.name like ('%' || $query || '%') or texts.desc like ('%' || $query || '%'))`,
      {
        $num: possibleValueNumber,
        $query: value,
      },
    );
    return results.map((obj) => plainToInstance(YGOProCard, obj));
  }
  private randomCardInDB(db: SQL.Database) {
    const allCardIds = this.querySQL(
      db,
      `select id from datas where ${commonFilter}`,
    );
    if (!allCardIds.length) return;
    const id: number = Random.pick(allCardIds).id;
    const [result] = this.queryInDB(id.toString(), db);
    return result;
  }
  private randomCard() {
    const db = Random.pick(this.dbs);
    if (!db) return;
    return this.randomCardInDB(db);
  }
  private queryInAllDBs(value: string, matchCount?: number) {
    return _.sortBy(
      _.uniqBy(
        _.flatten(this.dbs.map((db) => this.queryInDB(value, db))),
        (c) => c.id,
      ),
      (c) => c.getDistanceFrom(value),
    ).slice(0, matchCount || this.config.matchCount);
  }
  private async loadDBs(SQL: SQL.SqlJsStatic) {
    const allDatabaseFiles = _.flatten(
      await Promise.all(
        this.config.databasePaths.map(async (p) =>
          (await fs.readdir(p))
            .filter((f) => f.endsWith('.cdb'))
            .map((f) => path.join(p, f)),
        ),
      ),
    );
    this.logger.info(
      `Will load cards from ${allDatabaseFiles.length} databases.`,
    );
    const buffers = await Promise.all(
      allDatabaseFiles.map((f) => fs.readFile(f)),
    );
    this.dbs = buffers.map((b) => new SQL.Database(b));
  }

  async onConnect() {
    const SQL = await initSqlJs();
    return this.loadDBs(SQL);
  }

  onDisconnect() {
    this.dbs.forEach((db) => db.close());
  }

  async renderCard(card: YGOProCard) {
    if (!this.config.usePuppeteer || !this.puppeteer) {
      return card.getDisplayString(this.config);
    }
    return (
      <html>
        <body>
          <div>
            <image src={card.getPic(this.config)} />
          </div>
          <div>
            {card.name} {card.id}
          </div>
          <div>
            <p>{card.getDescription(this.config)}</p>
          </div>
        </body>
      </html>
    );
  }

  @UseCommand('{{commandName}} [query]')
  @CommandLocale('zh', localeZh.cardCommand)
  @CommandLocale('en', localeEn.cardCommand)
  async onCardCommand(
    @PutArg(0) query: string,
    @PutOption('random', '--random') random: boolean,
    @PutOption('count', '-c <count:posint>') count: number,
    @PutSession() session: Session,
  ) {
    if (!query && !random) {
      await session.send(session.text('.prompt-input'));
      query = await session.prompt();
    }
    if (!query && !random) return;
    if (!count) {
      count = this.config.matchCount;
    }
    const cards = random
      ? [this.randomCard()]
      : this.queryInAllDBs(query, count);
    if (!cards.length) {
      return session.text('.not-found');
    }
    if (cards.length === 1) {
      return this.renderCard(cards[0]);
    }
    const itemLines = cards.map((c, i) => `${i + 1}. ${c.getIdAndName()}`);
    const borderLength = Math.max(...itemLines.map((l) => l.length)) + 1;
    await session.send(
      session.text('.select-prefix-part1') +
        query +
        session.text('.select-prefix-part2') +
        '\n' +
        _.repeat('-', borderLength) +
        '\n' +
        itemLines.join('\n') +
        '\n' +
        _.repeat('-', borderLength) +
        '\n' +
        session.text('.select-suffix'),
    );
    const reply = await session.prompt();
    if (!reply) {
      return;
    }
    const replyNumber = parseInt(reply);
    const card =
      cards[replyNumber - 1] ||
      cards.find((c) => c.id === replyNumber) ||
      cards.find((c) => c.name === reply) ||
      cards.find((c) => c.name.includes(reply)) ||
      cards.find((c) => c.desc.includes(reply));
    if (!card) {
      return session.text('.not-found');
    }
    return this.renderCard(card);
  }
}
